/* Copyright 2021 Tom A. Wagner <tom.a.wagner@protonmail.com>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as published by
   the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   SPDX-License-Identifier: GPL-3.0-only
*/

@define-color media-type-audio   rgb( 50, 100, 240);
@define-color media-type-video   rgb(200, 200,   0);
@define-color media-type-midi    rgb(200,   0,  50);
@define-color media-type-unknown rgb(128, 128, 128);

.audio {
    background: @media-type-audio;
    color: black;
}

.video {
    background: @media-type-video;
    color: black;
}

.midi {
    background: @media-type-midi;
    color: black;
}

node {
    /* Compared to the default card color, this is not transparent in dark-mode
       and provides a better contrast to the background in light mode */
	background-color: @headerbar_bg_color;
}

node .node-title {
	padding: 4px 7px;
}

port label {
    padding: 4px 6px;
}

port-handle {
    border-radius: 50%;
    background-color: @media-type-unknown;
}

button.rounded {
    padding: 6px;
    border-radius: 9999px;
}

entry.rounded {
    border-radius: 9999px;
}

entry.rounded > :first-child {
    padding-left: 12px;
}

entry.rounded > :nth-child(2) {
    padding-right: 12px;
}
